#!BPY

"""
Name: 'Bitmap Font Creator'
Blender: 245
Group: 'Image'
"""

############################################################
#                                                          #
# (C) 2008  Zdeno Ash Miklas                               #
#                                                          #
# Bitmap Font Creator                                      #
#                                                          #
############################################################


import Blender
from Blender.BGL import *
from Blender.Draw import *
from PIL import Image, ImageFont, ImageDraw


# font class
class Font:
  # constructor
  def __init__(self, size, fontFile, baseSet, charMap, padding, stdLet, delLet):
    # font attributes
    self.size = size
    self.padding = padding
    # open font
    self.font = ImageFont.truetype(fontFile, size, 0, 'unic')
    # font parameters
    self.bottomLine = self.rendStr(stdLet).getbbox()[3]
    self.delLetter = delLet
    delLetBox = self.rendStr(delLet + delLet).getbbox()
    self.delWidth = int((delLetBox[2] - delLetBox[0]) / 2)
    # create character map
    self.chars = [['\0', u'@']]
    # basic chars
    if baseSet:
      for fontChar in range(ord(' '), 0x7F):
        self.chars.append([chr(fontChar), unichr(fontChar)])
    # special characters
    if len(charMap) > 0:
      mapFile = file(charMap, 'r')
      try:
        for mapLine in mapFile:
          if len(mapLine) > 0:
            try:
              uniCode = int(mapLine[1:], 16)
              if uniCode > 0:
                self.chars.append([mapLine[0], unichr(uniCode)])
            except: pass
      finally:
        mapFile.close()
    # sort characters
    self.chars.sort()
    # create font characters
    for charIdx in range(len(self.chars)):
      self.chars[charIdx] = FontChar(self.chars[charIdx], self)

  # build bitmap font
  def build(self, width):
    # size of bitmap font image
    imgSize = [calcSize(width), 0]
    # header row count
    self.headRow = int((12 + len(self.chars) * 12 + width - 1) / width)
    # row to store char images
    charRow = 1
    # place char images to row
    rowIdx = 0
    while rowIdx < len(self.chars):
      # char row parameters
      idx = rowIdx
      actWidth = self.padding + 1
      maxAbove = 0
      maxBelow = 0
      # check chars till the end of row
      while idx < len(self.chars) and actWidth + self.chars[idx].size[0] + self.padding < width:
        # add char width and padding
        actWidth += self.chars[idx].size[0] + 2 * self.padding + 1
        # check char sizes
        if maxAbove < self.chars[idx].above:
          maxAbove = self.chars[idx].above
        if maxBelow < self.chars[idx].size[1] - self.chars[idx].above:
          maxBelow = self.chars[idx].size[1] - self.chars[idx].above
        # next char
        idx += 1
      # set char positions
      col = 1
      for i in range(rowIdx, idx):
        # set char position
        self.chars[i].pos[0] = col
        self.chars[i].pos[1] = charRow + maxAbove - self.chars[i].above
        # get to position of the next char
        col += self.chars[i].size[0] + 2 * self.padding + 1
      # go to next row
      charRow += maxAbove + maxBelow + 2 * self.padding + 1
      rowIdx = idx
    # set image height
    imgSize[1] = calcSize(charRow) + self.headRow
    if charRow + self.headRow > imgSize[1]:
      imgSize[1] = calcSize(2 * charRow) + self.headRow
    # create image
    self.image = Image.new('L', imgSize, 0)
    # initial data
    pos = [0, 0]
    for char in 'BFNT\0\0':
      self.putPix(pos, ord(char), 1)
    # chars count
    self.putPix(pos, len(self.chars), 2)
    # font size
    self.putPix(pos, self.size, 2)
    # fill zeros
    self.putPix(pos, 0, 2)
    # write all font characters
    for item in self.chars:
      item.write(pos)

  # render string
  def rendStr (self, str, pos = None):
    # render image size
    size = 256
    # if position is not specified, set default
    if pos is None:
      posVal = int((size - self.size) / 2)
      pos = (posVal, posVal)
    # create image
    img = Image.new('L',(size,size), 0)
    drw = ImageDraw.Draw(img)
    # render string
    drw.text(pos, str, font = self.font, fill = 255)
    # return render
    return img

  # put value to pixels
  def putPix (self, pos, val, size):
    # write all bytes of value (starting from MSB)
    for idx in range(size - 1, -1, -1):
      # write byte
      self.image.putpixel(pos, (val >> (8 * idx)) & 0xff)
      # move to next pixel
      pos[0] += 1
      # if on the end of line
      if pos[0] >= self.image.size[0]:
        # next line
        pos[0] = 0
        pos[1] += 1


# class for font character
class FontChar:
  # constructor
  def __init__(self, code, font):
    # character code and font reference
    self.code = code
    self.font = font
    # char position
    self.pos = [0, 0]
    # char width
    img = font.rendStr(font.delLetter + code[1] + font.delLetter)
    box = img.getbbox()
    box = (box[0] + font.delWidth, 0, box[2] - font.delWidth, img.size[1])
    # get char image
    img = img.crop(box)
    self.width = img.size[0]
    box = img.getbbox()
    # if char image is available
    if box is not None:
      # add 1 pixel on each side
      if box[0] > 0: box = (box[0] - 1, box[1], box[2], box[3])
      if box[1] > 0: box = (box[0], box[1] - 1, box[2], box[3])
      if box[2] < img.size[0]: box = (box[0], box[1], box[2] + 1, box[3])
      if box[3] < img.size[1]: box = (box[0], box[1], box[2], box[3] + 1)
      self.image = img.crop(box)
      # rest of char attributes
      self.size = [box[2] - box[0], box[3] - box[1]]
      self.before = box[0]
      self.above = font.bottomLine - box[1]
    # if character is empty
    else:
      self.image = img.crop((0, 0, self.width, 1))
      self.size = [self.width, 1]
      self.before = 0
      self.above = 0

  # write char to font bitmap
  def write (self, pos):
    # char code
    code = ord(self.code[0])
    # hack to workaround Blender's bug
    #if code >= 0x80: code = code | 0xFF00
    # write attributes to bitmap header
    self.font.putPix(pos, code, 2)
    for co in self.pos:
      self.font.putPix(pos, co, 2)
    self.font.putPix(pos, self.before + self.font.padding, 1)
    self.font.putPix(pos, self.above + self.font.padding, 1)
    for co in self.size:
      self.font.putPix(pos, co + 2 * self.font.padding, 1)
    self.font.putPix(pos, self.width + 2 * self.font.padding, 1)
    self.font.putPix(pos, 0, 1)
    # write char image
    self.font.image.paste(self.image, (self.pos[0] + self.font.padding,
      self.pos[1] + self.font.headRow + self.font.padding,
      self.pos[0] + self.size[0] + self.font.padding,
      self.pos[1] + self.size[1] + self.font.headRow + self.font.padding))


# calculate size of image
def calcSize(size):
  while size & (size - 1) != 0: size = size & (size - 1)
  return size

#draw GUI
def draw ():
  glClearColor(0.6, 0.6, 0.6, 1.0)
  glClear(GL_COLOR_BUFFER_BIT)
  Button("Create Bitmap Font",10,10,10,128,16)

  global name, size, fontFold, fontFile, mapFile, width, padding
  global baseSet, setObj, basicChar, delimiter
  name = String('Font File: ',11,10,150,192,16,name.val,32,'Bitmap font texture name')
  size = Number('Font size: ',11,206,150,128,16,size.val,4,128,'Bitmap font size')
  Button('>',12,10,130,16,16)
  fontFold = String('Font Folder: ',14,26,130,384,16,fontFold.val,128,'Bitmap font folder')
  Button('>',13,10,110,16,16)
  fontFile = String('TTF File: ',16,26,110,384,16,fontFile.val,128,'Source TTF file')
  Button('>',15,10,90,16,16)
  mapFile = String('Map File: ',16,26,90,384,16,mapFile.val,128,'Character mapping file')
  width = Number('Width: ',17,10,70,128,16,width.val,64,1024,'Bitmap width')
  padding = Number('Padding: ',17,142,70,128,16,padding.val,0,32,'Padding for characters')
  baseSet = Toggle('Use basic characters',20,10,50,128,16,baseSet.val,'Include basic ASCII character set in font')
  setObj = Toggle('Setup selected object',20,142,50,128,16,setObj.val,'Setup selected object to use bitmap font')
  basicChar = String('Base Char: ',18,10,30,128,16,basicChar.val,1,'Base character to measure font properties')
  delimiter = String('Delimiter: ',19,142,30,128,16,delimiter.val,1,'Delimiter character')

# event handling
def event (ev, val):
  if ev == ESCKEY: Exit()


# event processing
def proc (ev):
  global name, size, fontFold, fontFile, mapFile, width, padding
  global baseSet, setObj, basicChar, delimiter
  # file select events
  if ev == 12:
    Blender.Window.FileSelector(setFontFold, 'Select font folder', fontFold.val)
  if ev == 13:
    Blender.Window.FileSelector(setFontFile, 'Select TTF file', fontFile.val)
  if ev == 15:
    Blender.Window.FileSelector(setMapFile, 'Select map file', mapFile.val)
  # check folder
  if ev == 14:
    setFontFold(fontFold.val)
  # bitmap width
  if ev == 17:
    width.val = calcSize(width.val)
  # create font bitmap
  if ev == 10:
    fnt = Font(size.val, fontFile.val, baseSet.val, mapFile.val, padding.val,
      basicChar.val, delimiter.val)
    fnt.build(width.val)
    fnt.image.save(fontFold.val + name.val)
    try:
      fntImg = Blender.Image.Get(name.val)
      if fntImg is not None: fntImg.reload()
    except:
      fntImg = Blender.Image.Load(fontFold.val + name.val)
    # check, if active object can be mapped to font
    if setObj.val:
      obj = Blender.Object.GetSelected()[0]
      if obj.getType() == 'Mesh':
        msh = obj.getData(False, True)
        if len(msh.faces) == 1 and len(msh.verts) == 4:
          # set mesh to face mode
          msh.faceUV = True
          # set font to object
          uv = [[float(fnt.chars[0].pos[0])/fnt.image.size[0],
            1.0 - float(fnt.chars[0].pos[1])/(fnt.image.size[1] - fnt.headRow)],
            [float(fnt.chars[0].size[0] + 2 * fnt.padding)/fnt.image.size[0],
            float(fnt.chars[0].size[1] + 2 * fnt.padding)/(fnt.image.size[1] - fnt.headRow)]]
          face = msh.faces[0]
          face.mode = face.mode | Blender.Mesh.FaceModes.TEX | Blender.Mesh.FaceModes.TEXT
          face.uv = (Blender.Mathutils.Vector(uv[0][0] + uv[1][0], uv[0][1]),
            Blender.Mathutils.Vector(uv[0][0], uv[0][1]),
            Blender.Mathutils.Vector(uv[0][0], uv[0][1] - uv[1][1]),
            Blender.Mathutils.Vector(uv[0][0] + uv[1][0], uv[0][1] - uv[1][1]))
          face.image = fntImg
          face.transp = Blender.Mesh.FaceTranspModes.ALPHA
          # set Text property
          try:
            textProp = obj.getProperty('Text')
          except:
            obj.addProperty('Text', 'Lorem ipsum', 'STRING')
   # redraw GUI
  Blender.Window.RedrawAll()

def setFontFold (fileName):
  global fontFold
  fontFold.val = fileName
  if len(fontFold.val) > 0 and fontFold.val[-1] != '/' and fontFold.val[-1] != '\\':
    fontFold.val += '/'

def setFontFile (fileName):
  global fontFile
  fontFile.val = fileName

def setMapFile (fileName):
  global mapFile
  mapFile.val = fileName

# main program
if __name__ == "__main__":
  # initialize GUI objects
  global name, size, fontFold, fontFile, mapFile, width, padding
  global baseSet, setObj, basicChar, delimiter
  name = Create('')
  size = Create(32)
  fontFold = Create('')
  fontFile = Create('')
  mapFile = Create('')
  width = Create(512)
  padding = Create(0)
  baseSet = Create(True)
  setObj = Create(False)
  basicChar = Create('A')
  delimiter = Create('|')

  # register event handler
  Register(draw, event, proc)

  # draw GUI
  Blender.Redraw()